"""
This is a non-standard CRC check for the Baler device. Accept it as a
black box!
"""


class CRC:
    """
    CRC check for the Baler device.
    """
    TABLE = []
    MASK = 0xFFFFFFFF
    CRC_POLYNOMIAL = 1443300200

    @classmethod
    def create_table(cls):
        """
        The docstring
        """
        cls.TABLE = [None] * 256
        # t_data = 0
        # accum = 0
        for count in range(0, 256):
            t_data = (count << 24)
            accum = 0
            for _ in range(1, 9):
                if (t_data ^ accum) & 2 ** 31:
                    accum = (accum << 1) ^ cls.CRC_POLYNOMIAL
                    accum = accum & cls.MASK
                else:
                    accum = (accum << 1) & cls.MASK
                t_data = (t_data << 1) & cls.MASK
            cls.TABLE[count] = accum

    @classmethod
    def compute_crc(cls, buf):
        """
        Compute the CRC value.

        Keyword arguments:
            buf -- bytes buffer (default)
        Returns:
            new_crc: int compute CRC value.
        """
        if not cls.TABLE:
            cls.create_table()
        lng = 0
        first_byte = 0
        data_ptr = 0
        data_len = len(buf)
        new_crc = None
        while data_len > 0:
            this_byte = buf[data_ptr]
            new_crc = cls.MASK & (lng << 8) ^ cls.TABLE[(first_byte ^
                                                         this_byte) & 255]
            first_byte = new_crc >> 24
            lng = new_crc
            data_ptr += 1
            data_len -= 1
        return new_crc
