"""
Give Quanterra Baler-14 ip address
Original Author: Lloyd Carothers IRIS/PASSCAL
Revised: Omid Hosseini IRIS/PASSCAL
"""
import glob
import logging
import socket
import struct
import subprocess
import sys
from glob import glob
from ipaddress import IPv4Network, ip_interface
from sys import stdout
from time import sleep
from typing import Set

from pexpect import spawn
from psutil import net_if_addrs

from . import tag_config
from .bal_packet import Packet
from .util import check_tags, user_decision_tags

VERSION = '2021.000'

# noinspection PyArgumentList
logger = logging.getLogger("baler")
logger.debug('sys.exec: %s', sys.executable)
logger.debug('prog: %s', __file__)
logger.info('version: %s', VERSION)


if not hasattr(socket, 'IP_PKTINFO'):
    if sys.platform == 'darwin':
        socket.IP_PKTINFO = 26
    if sys.platform == 'linux' or sys.platform == 'win':
        socket.IP_PKTINFO = 8


def socket_baler_listen():
    """
    Return a socket listening for baler ready packets.
    host_ip is equivalent to address ip for bind.
    '' = INADDR_ANY, or any interface.
    If this is used and there are multiple interfaces on this host,
    baler ready commands will be heard, but you will not know which interface
    to send the broadcast response to.
    """
    host = ''
    port = 65535
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_PKTINFO, 1)
    # try to bind to interface
    # this did not work
    # iface = str('enx8cae4cf4fae1' + '\0').encode()
    # sock.setsockopt(socket.SOL_SOCKET, socket.SO_BINDTODEVICE, iface)
    sock.bind((host, port))
    return sock


def socket_com(host):
    """
    TODO: Add Docstring!
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.settimeout(2)
    sock.bind((str(host), 0))
    return sock


def dev_from_msgs(msgs):
    """
    TODO: Add Docstring!
    """
    for msg in msgs:
        if msg[1] == socket.IP_PKTINFO:
            if_index = struct.unpack('I8B', msg[2])[0]
            dev = socket.if_indextoname(if_index)
            return dev
    return None


def ip_from_dev(dev):
    """
    TODO: Add Docstring!
    """
    addresses = [ad for ad
                 in net_if_addrs()[dev]
                 if ad.family is socket.AF_INET]
    logger.debug("%s 'has IPs: [%s]", dev, '. '.join(map(str, addresses)))
    if len(addresses) == 0:
        logger.error("Dev: %s no address.", dev)
        raise Exception("Dev has no address.")
    if len(addresses) > 1:
        logger.error("Dev: %s doesn't have a single IP:%s", dev, addresses)
        raise Exception("Dev doesn't have a single IP address")
    baler_addr = addresses[0]
    net = IPv4Network((baler_addr.address, baler_addr.netmask),
                      strict=False)
    ip_if = ip_interface((baler_addr.address, net.prefixlen))
    return ip_if


def is_ip_free(ip_addr):
    """
    TODO: Add Docstring!
    """
    ip_addr = str(ip_addr)
    try:
        proc = subprocess.run(['ping', '-c1', ip_addr], timeout=0.3,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                              check=True)
    except subprocess.TimeoutExpired:
        return True
    return not proc.returncode == 0


def next_free_ip(host):
    """
    TODO: Add Docstring!
    """
    count = 0
    for free_ip in (i for i in host.network.hosts() if i is not host):
        if is_ip_free(free_ip):
            return free_ip
        count += 1
        if count > 100:
            return None
    return None


def address_balers(stop, tag=None):
    """
    If Tag is set, will only address that Balers, otherwise address all balers.

    keyword arguments:
    tag -- (tuple) list of all tags for addressing.
    """
    chk_tags = None
    tag_ids: Set[int] = set()
    ident_tags: Set[int] = set()
    if tag is not None and check_tags(tag) is False:
        logger.error("Incorrect input format for tag_ids")
        raise Exception("Incorrect input for the tag_ids")
    if tag is not None and check_tags(tag) is True:
        # Make sure all input tags have int value.
        tag_ids = set(int(i) for i in tag)
        # Get the number of passed tags. The number of added devices can't be
        # more than the number of tags that the user provides.
        chk_tags = len(tag_ids)
        count = 0

    sock = socket_baler_listen()
    count = 0
    while True:
        if stop():
            logger.info("Server is being shutdown.")
            break
        logger.info('Listening for balers...')
        try:
            sock.settimeout(20)
            raw, msgs, flags, addr = sock.recvmsg(44, 128)
            baler_port = addr[1]
            dev = dev_from_msgs(msgs)
            logger.info('Received baler (flag = %d) ready packet from %s on '
                         'dev: %s', flags, addr, dev)
            brdy_pkt = Packet()
            brdy_pkt.unpack(raw)
            logger.debug(brdy_pkt)
        except socket.timeout as err:
            continue

        if tag and brdy_pkt.q330_sn2:
            logger.info('Found TAG: %s adding to the list', brdy_pkt.q330_sn2)
            ident_tags.add(int(brdy_pkt.q330_sn2))
            count += 1
        localhost = ip_from_dev(dev)
        baler_ip = str(next_free_ip(localhost))

        # Keep track of all connected Balers.
        tag_config.set_tag(brdy_pkt.q330_sn2, str(baler_ip), baler_port)

        vack = Packet()
        vack.pack_vack(baler_ip, localhost.netmask, brdy_pkt)
        logger.debug(vack)
        vack.crc_check()
        sock_com = socket_com(localhost.ip)
        bytes_sent = sock_com.sendto(vack.buf, ('255.255.255.255', baler_port))
        # It looks like the computer can get ahead of the baler and send
        # commands too quickly before the baler has finished doing something
        # and is ready to respond to a sent command (e.g. time and ping).
        # This pause for sure improved things on Macs. There are still random
        # response failures, but it looks like
        # the IP setting still happens. When something like this fails with
        # BaleAddr you still can't talk to the baler.
        sleep(1.0)
        logger.info("Sent vack %d of %d", bytes_sent, len(vack.buf))
        # Set the time.
        time_pkt = Packet()
        time_pkt.pack_set_baler_time(brdy_pkt)
        # This works, but if you have more than one device with the same
        # networks You still won't be able to route back to one of them i.e.
        # no web page or download bytes_sent = sock_com.sendto(time_pkt.buf,
        # ('255.255.255.255', baler_port))
        bytes_sent = sock_com.sendto(time_pkt.buf, (baler_ip, baler_port))
        sleep(1.0)
        logger.info('Sent time %d of %d', bytes_sent, len(time_pkt.buf))

        try:
            raw, addr = sock_com.recvfrom(14)
            logger.info('Received time ack from %s %d of %d',
                         addr, len(raw), 14)
            time_ack_pkt = Packet()
            time_ack_pkt.unpack(raw)
            logger.debug(time_ack_pkt)
            logger.info('TAG: %s <=> IP: %s', brdy_pkt.q330_sn2, baler_ip)
            print(f'Visit: http://{baler_ip}')
        # Other exceptions may be caught.
        except socket.error as err:
            logger.error("Didn't received response from time cmd.")
            print(err)
            logger.error(err)

        # TODO: TEST PING BALER MUST BE MOVED TO A SEPARATE FUNCTION.
        ping = Packet()
        ping.pack_ping()
        bytes_sent = sock_com.sendto(ping.buf, (baler_ip, baler_port))
        logger.info('Sent ping %d of %d', bytes_sent, len(ping.buf))
        logger.info('sent to %s: %d' % (baler_ip, baler_port))
        sleep(1.0)
        try:
            raw, addr = sock_com.recvfrom(14)
            logger.info('Received ping from %s %s of %d', addr, len(raw), 14)
            ack_pkt = Packet()
            ack_pkt.unpack(raw)
            logger.debug(ack_pkt)
            logger.info('TAG: %s <=> IP: %s is ALIVE', brdy_pkt.q330_sn2,
                         baler_ip)
        # Other exceptions may be caught.
        except socket.error as err:
            logger.error("Didn't received response from ping cmd.")
            print(err)
            logger.error(err)

        # Input tags are received and processed by address_baler() function
        if count == chk_tags:
            missing_tags = tag_ids - ident_tags
            extra_tags = ident_tags - tag_ids
            # Everything is fine and all tags are identified. Return True.
            if tag_ids == ident_tags or len(missing_tags) == 0:
                logger.info("All tags %s are found.", tag_ids)
                print("The following tags are addressed:")
                tag_config.print_tags()
                return True

            # There are issues related to the input tags. Addressing them here.
            usr_dec = user_decision_tags(missing_tags, extra_tags)
            print(usr_dec)
            # The user wants to add Balers with missing tags. continue
            # accepting Balers until all tags are added.
            if usr_dec["conn_miss"]:
                count -= len(missing_tags)
                print("Inside the conn_miss and the count is %d" % count)
            # The user doesn't want to keep the extra identified tags. Shut
            # those Balers down.
            if not usr_dec["kpt_extra"]:
                print("Inside the kpt_extra")
                print(extra_tags)
                tag_config.print_tags()
                for tgs in extra_tags:
                    print("Inside kep_extra and the tags is: %d" % tgs)
                    baler_shutdown(sock_com, tuple(tag_config.get_tag(tgs)))
                    # The Baler is shutdown so remove it from the identified
                    # tags and tag_config dictionary
                    ident_tags.remove(tgs)
                    tag_config.delete_tag(tgs)


def clean_baler(sock_com, baler_ip_port):
    """
    TODO: Add Docstring!
    """
    clean_p = Packet()
    clean_p.pack_clean()
    bytes_sent = sock_com.sendto(clean_p.buf, baler_ip_port)
    logger.info('Sent clean %d of %d', bytes_sent, len(clean_p.buf))


DEV_FORMAT = '/proc/sys/net/ipv4/conf/{0}/rp_filter'


def rp_filter_ok(dev):
    """
    TODO: Add Docstring!
    """
    if not dev.startswith('/'):
        dev = DEV_FORMAT.format(dev)
    with open(dev, 'r') as f_path:
        status = int(f_path.read())
        logger.debug('%s = %s', dev, status)
        return status == 0


def check_rp_filters_off(dev=None):
    """
    Tests if linux rp_filter kernel option is off.
    Option must be off to get baler ready broadcast packets.
    dev is network device e.g. enp1s0.
    if dev None checks all
    """
    if sys.platform != 'linux':
        logger.debug('Platform not linux, should be fine.')
        return True
    # Check 'global' kernel option
    if not rp_filter_ok('all'):
        dev_path = DEV_FORMAT.format('all')
        logger.error('Global rp_filter set. Can NOT get baler discovery '
                     'packets from linux kernel. to fix: sudo bash -c "echo 0 '
                     '%s"', dev_path)
        return False
    devs = []
    if dev is not None:
        devs.append("all")
        devs.append(dev)
    else:
        devs = glob('/proc/sys/net/ipv4/conf/*/rp_filter')
    for net_dev in devs:
        status = rp_filter_ok(dev)
        if status:
            logger.info('Fixed. Can listen on %s', net_dev)
        else:
            stdout.write("Executing: sudo bash -c \"echo 0 > %s\"\n" % net_dev)
            stdout.write("   Enter the password")
            stdout.flush()
            try_spawn = spawn("sudo bash -c \"echo 0 > %s\"" % devs)
            try_spawn.timeout = 2
            try_spawn.expect("password")
            try_spawn.interact()
            # logging_setup.info('Can NOT listen on %s To listen on this '
            #              'device: : sudo bash -c "echo 0 %s', dev, dev)
    return any((rp_filter_ok(dev) for dev in devs))


def baler_shutdown(sock_com, bal_ip_port):
    """
    Shutdown baler using the ip address and port number.
    """
    shutdown_packet = Packet()
    shutdown_packet.pack_shutdown()
    bytes_sent = sock_com.sendto(shutdown_packet.buf, bal_ip_port)
    logger.info('Sent shutdown %d of %d', bytes_sent, len(shutdown_packet.buf))


def baler_disconnect_all(baler_socket):
    """
    Disconnect all connected Balers.

    keyword arguments:
    baler_socket -- (socket) Socket used for Baler connection.
    """
    tags = tag_config.retrieve_all_tags()
    for tgs in tags:
        baler_shutdown(baler_socket, (tgs[0], tgs[1]))
        sleep(1)
    logger.info("All Balers are disconnected.")
    return 0
