"""
This Module acts as a singleton class in a server mode and keep track of all
connected Baler devices.
Functionalities:
set_tag(bal_tag, bal_ip, bal_port)
get_tag(bal_tag)
delete_tag(bal_tag)
print_tags()
tag_exist(bal_tag)
"""


BALER_TAGS = dict()


def set_tag(bal_tag, bal_ip, bal_port):
    """
    Add a tag with associated ip and port to the BALER_TAGS

    keyword arguments:
    tag -- (int) Baler's tag ID.
    ip -- (str) IP address assigned to the Baler with the tag ID.
    port -- (int) Port assigned to the Baler with the tag ID.
    """
    if bal_tag in BALER_TAGS.keys():
        print("The tag, %d, with IP address: %s and Port: %d already exists "
              "in the list", bal_tag, str(bal_ip), bal_port)
    else:
        BALER_TAGS[bal_tag] = ([str(bal_ip), bal_port])


def get_tag(bal_tag):
    """
    Return the Baler's IP address and port as list

    keyword arguments:
    bal_tag -- (int) The Baler tag id
    """
    return BALER_TAGS[bal_tag]


def delete_tag(bal_tag):
    """
    Delete the Baler from the list.

    keyword arguments:
    bal_tag -- (int) The Baler tag id
    """
    removed_tag = BALER_TAGS.pop(bal_tag)
    print("The key: %s is removed", removed_tag)


def print_tags():
    """
    Print the content of the Baler tag list
    """
    for key, value in BALER_TAGS.items():
        print(key, ' : ', value)


def tag_exist(bal_tag):
    """
    Check and see if a given tag exists in the list.

    keyword arguments:
        bal_tag -- (int) The Baler tag id
    """
    return True if bal_tag in BALER_TAGS.values() else False


def retrieve_all_tags():
    """
    Return all tags and their IP address and port
    """
    return BALER_TAGS
