"""
This module is used to setup root logging_setup configuration,
If LOG_CFG variable is set, then use it for logging_setup configuration path.
Since we are using yaml configuration, we need yaml module to load
    configuration file.
Set Root logger configuration using `logging_setup.config.dictConfig()`.
Any exception results in setting up root logger in default configuration.
"""
import logging
import logging.config
import os
import sys
import yaml

from baler_addr.util import mkdir_p


def setup_logging(default_name='logging.config', default_level=logging.INFO,
                  env_key='LOG_CFG'):
    """
    Logging Setup function

    keyword arguments:
    default_name -- (str) Logging configuration file name
    default_level -- (int) Default logging_setup level
    env_key -- (str) Logging config path set in environment variable
    """
    # Create the LOG directory if needed.
    mkdir_p("bline/logging_setup/LOG")
    __location__ = os.path.realpath(
        os.path.join(os.getcwd(), os.path.dirname(__file__)))
    path = os.path.join(__location__, default_name)
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, 'rt') as cnf:
            try:
                config = yaml.safe_load(cnf.read())
                logging.config.dictConfig(config)
            except Exception as err:
                print('Error in Logging Configuration. Using default configs',
                      err)
                logging.basicConfig(level=default_level, stream=sys.stdout)
    else:
        logging.basicConfig(level=default_level, stream=sys.stdout)
        print('Failed to load configuration file. Using default configs')
